package envelope

import (
	"fmt"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"
)

// BuilderFromFragment builds an Envelope from a serie of api.Envelope
// fragments.
type BuilderFromFragment struct {
	builder  *Builder
	sorter   *tools.EnvelopeFragmentSorter
	messages map[api.UUID]*StreamMessage

	sorted []*api.Envelope

	Complete bool
	done     chan struct{}
}

// Envelope returns the envelope being built
func (eb *BuilderFromFragment) Envelope() Envelope {
	return eb.builder
}

// Done returns a channel that will be closed with the reception of fragments is complete
func (eb *BuilderFromFragment) Done() <-chan struct{} {
	return eb.done
}

// dispatch dispatches the next fragment into the envelope
func (eb *BuilderFromFragment) dispatch(fragment *api.Envelope) error {
	if len(eb.messages) == 0 {
		for _, id := range eb.sorter.ExpectedEvents {
			ev := eb.sorter.Events[id]
			message := NewStreamMessage(ev.EventID, ev.Type)
			if err := eb.builder.Add(message); err != nil {
				return err
			}
			eb.messages[ev.EventID] = message
		}
		if err := eb.builder.Close(); err != nil {
			return err
		}
	}
	for _, eventFragment := range fragment.Events {
		message, ok := eb.messages[api.MustUUIDFromBytes(eventFragment.ID)]
		if !ok {
			return fmt.Errorf("received event with unexpected id: %s", eventFragment.GetIDAsUUID())
		}
		if message.Checksum() == 0 && eventFragment.Checksum != 0 {
			message.SetChecksum(api.Checksum(eventFragment.Checksum))
		}
		if len(eventFragment.Items) != 0 {
			if err := message.addChunks(eventFragment.Items); err != nil {
				return err
			}
		}
	}
	return nil

}

// Add adds a fragment
func (eb *BuilderFromFragment) Add(fragment *api.Envelope) error {
	eb.sorter.Add(fragment)
	for eb.sorter.Next() {
		next, err := eb.sorter.Get()
		if err != nil {
			return err
		}
		eb.sorted = append(eb.sorted, next)
	}
	if eb.sorter.EventTypesKnown {
		for _, fragment := range eb.sorted {
			if err := eb.dispatch(fragment); err != nil {
				_ = eb.Cancel(err)
				return err
			}
		}
		eb.sorted = eb.sorted[0:0]
	}
	if eb.sorter.Complete {
		for _, message := range eb.messages {
			if err := message.addChunks(nil); err != nil {
				return err
			}
		}
		eb.Complete = true
		close(eb.done)
	}

	return nil
}

// Cancel stops builder and set an error on the envelope
func (eb *BuilderFromFragment) Cancel(err error) error {
	select {
	case <-eb.done:
	default:
		close(eb.done)
	}
	return eb.builder.CloseWithError(err)
}

// NewFromFragments returns a Envelope that can be fed with a stream
// of envelope fragments. After all fragments are received, the functions returns io.EOF
func NewFromFragments(envelopeID api.UUID) *BuilderFromFragment {
	builder := BuilderFromFragment{
		sorter:   tools.NewEnvelopeFragmentSorter(),
		messages: make(map[api.UUID]*StreamMessage),
		builder:  NewOpenedEnvelopeWithID(envelopeID),
		done:     make(chan struct{}),
	}

	return &builder
}
